#ifndef DEVDLS8515_H
#define DEVDLS8515_H

#include "tyLib.h"

#include <time.h>

#include <epicsTypes.h>
#include <epicsStdio.h>
#include <epicsString.h>
#include <epicsExit.h>
#include <epicsStdio.h>
#include <epicsEvent.h>
#include <epicsMutex.h>
#include <epicsMessageQueue.h>
#include <cantProceed.h>
#include <ellLib.h>
#include <epicsExport.h>
#include <registryFunction.h>

#define DRIVER		"DLS8515"
#define CHANNELS	8
#define DEVNAMEMAX	64		/* maximum length of device name */
#define RXBUFFSIZE	32768		/* vxWorks device RX buffer size */
#define TXBUFFSIZE	256		/* vxWorks device TX buffer size */
#define XON		0x11
#define XOFF		0x13
#define FIFOSIZE	64
#define RXFIFOTRIG	56		/* RX FIFO trigger size */
#define TXFIFOTRIG	1		/* TX FIFO trigger size */
#define QUEUE		4		/* number of message queue entries per channel */

#define RX_PRIO		3		/* RX process priority */
#define TX_PRIO		4		/* TX process priority */

/*
	error messages

	epicsMessageQueueReceive
	Rx Ring Buffer full
	Receiver Overrun
	epicsMessageQueueTrySend
*/

#define MSG_MQRX_BIT	0
#define MSG_RXFULL_BIT	1
#define MSG_RXOVER_BIT	2
#define MSG_MQTX_BIT	3

#define MSG_NUM		4		/* number of error message types */

#define MSG_MQRX	(1 << MSG_MQRX_BIT)
#define MSG_RXFULL	(1 << MSG_RXFULL_BIT)
#define MSG_RXOVER	(1 << MSG_RXOVER_BIT)
#define MSG_MQTX	(1 << MSG_MQTX_BIT)

#ifndef DEVDLS8515_C

static const char *messages[MSG_NUM] = { "epicsMessageQueueReceive", "Ring Buffer full", "Overrun error", "epicsMessageQueueTrySend"};

#endif

/* device structure */

typedef struct
{
	TY_DEV tyDev;
	BOOL created;			/* true if this device is created */
	
	int channel;
	int baud;
	int data;
	int stop;
	int parity;
	int flow;
	
	short txtrg, rxtrg;
	
	uint64_t txbytes, rxbytes;	/* total bytes sent and received */
	
	void *_HY8515;			/* points back to the HY8515 structure */
	
} TY_8515_DEV;

typedef volatile epicsUInt16 *regBase;

typedef struct
{
	ELLNODE node;
	
	uint16_t card;				/* VME slot * 10 * IP slot */
	uint16_t vector;
	uint16_t model;
	volatile uint16_t txint;		/* each bit indicates an interrupting channel */
	volatile uint16_t rxint;		/* each bit indicates an interrupting channel */
	
	regBase reg;				/* base address of IP module registers */
	epicsUInt16 *uart;			/* base address of IP module memory for UART registers */
						/* these are byte registers on 16 bit boundaries */
					
	TY_8515_DEV dev[CHANNELS];		/* device descriptors */

	epicsEventId work;			/* trigger TX process */
	epicsMutexId mutex;			/* limits access to all LCR registers */
	epicsMessageQueueId queue;		/* transfer rx data from ISR */

	volatile uint64_t rxints, txints;	/* interrupts per second */
	
	volatile uint16_t errors[CHANNELS+1];	/* bit pattern of current errors */
	
	volatile uint16_t rxread[CHANNELS];	/* max RX FIFO count */
	
} HY8515;

struct FIFO
{
	uint8_t channel, lsr;
};

/* baud rate and clock divisor */

typedef struct
{
	int rate;
	int divisor;
} BAUD;

/*
* baudTable is a table of the available baud rates, and the values to write
* to the UART's baud rate divisor {high, low} register. the formula is
*
* 14745600 / 4 / (16 * baudrate)	MCR bit 7 = 1	(/4)	8XMODE = 0
* 14745600 / 1 / (16 * baudrate)	MCR bit 7 = 0	(/1)	8XMODE = 0
* 14745600 / 4 / ( 8 * baudrate)	MCR bit 7 = 1	(/4)	8XMODE = 1
* 14745600 / 1 / ( 8 * baudrate)	MCR bit 7 = 0	(/1)	8XMODE = 1
*
* Xilinx clock is 125kHz
*
*/

static const BAUD baudTable [] =
{
	{1200,  192},
	{2400,   96},
	{4800,   48},
	{9600,   24},
	{19200,  12},
	{38400,   6},
	{57600,   4},
	{115200,  2},
	{230400,  1}
};

/* Synchronous read and writes */

epicsUInt16 sysIn16(void *port);
epicsUInt32 sysIn32(void *port);

void sysOut16(void *port, epicsUInt16 data);
void sysOut32(void *port, epicsUInt32 data);

/*
	some sync/async macros
*/

#define VME_READ16(addr)		*(volatile epicsUInt16 *)(addr)
#define VME_READ16_SYNC(addr)		sysIn16((addr))

#define VME_READ32(addr)		*(volatile epicsUInt32 *)(addr)
#define VME_READ32_SYNC(addr)		sysIn32((addr))

#define VME_WRITE16(addr, data)		*(volatile epicsUInt16 *)((addr)) = (data)
#define VME_WRITE16_SYNC(addr, data)	sysOut16((addr), (epicsUInt16) (data))

#define VME_WRITE32(addr, data)		*(volatile epicsUInt32 *)((addr)) = (data)
#define VME_WRITE32_SYNC(addr, data)	sysOut32((addr), (epicsUInt32) (data))

#define UART_READ_SYNC(a, c, r)		sysIn16((a) + 16 * (c) + (r))
#define UART_READ(a, c, r)		*(volatile epicsUInt16 *)((a) + 16 * (c) + (r))
#define UART_WRITE_SYNC(a, c, r, d)	sysOut16((a) + 16 * (c) + (r), (d))
#define UART_WRITE(a, c, r, d)		*(volatile epicsUInt16 *)((a) + 16 * (c) + (r)) = (d)

/*******************************************************************************
*
* Define memory offsets of the 8515 device control registers in A16 IP space.
* These are offsets into the XILINX area, not the UART space.
*
*******************************************************************************/

#define REG_INT0	(0x0)
#define REG_INT1	(0x1)
#define REG_INT2	(0x2)
#define REG_INT3	(0x3)
#define REG_TIMCTRL	(0x4)
#define REG_TIMER	(0x5)
#define REG_TIMLSB	(0x6)
#define REG_TIMMSB	(0x7)
#define REG_8xMODE	(0x8)
#define REG_DUPLEX	(0x9)
#define REG_RESET	(0xA)
#define REG_SLEEP	(0xB)
#define REG_DREV	(0xC)
#define REG_DVID	(0xD)
#define REG_DEVCTR	(0xE)

/*******************************************************************************
*
* Index into the UART configuration table. There is one of these tables for
* each of the eight channels on the card.
*
*******************************************************************************/

#define UA_RHR		0x00
#define UA_THR		UA_RHR
#define UA_DLL		UA_RHR
#define UA_DLM		0x01
#define UA_IER		UA_DLM
#define UA_ISR		0x02
#define UA_FCR		UA_ISR
#define UA_LCR		0x03
#define UA_MCR		0x04
#define UA_LSR		0x05
#define UA_MSR		0x06
#define UA_RS485	0x06
#define UA_SPR		0x07
#define UA_FCTR		0x08
#define UA_EFR		0x09
#define UA_TXCNT	0x0A
#define UA_TXTRG	UA_TXCNT
#define UA_RXCNT	0x0B
#define UA_RXTRG	UA_RXCNT
#define UA_XCHAR	0x0C
#define UA_XOFF1	UA_XCHAR
#define UA_XOFF2	0x0D
#define UA_XON1		0x0E
#define UA_XON2		0x0F

/* global interrupt register encoding */

#define INT_RXRDY	0x01	/* Priority 1 - Recieve Data Ready */ 
#define INT_RXTO	0x02	/* Priority 2 - Recieve Data Time-Out */ 
#define INT_TXRDY	0x03	/* Priority 3 - Transmit Empty */

/* define bits in the Interrupt Status Register (ISR) */

#define ISR_INTBITS	0x3F	/* Mask to Get Interrupt Source Bit Only */
#define ISR_LSRRDY	0x06	/* Priority 1 - LSR Interrupt */ 
#define ISR_RXRDY	0x04	/* Priority 2 - Recieve Data Ready */ 
#define ISR_RXTO	0x0C	/* Priority 3 - Recieve Data Time-Out */ 
#define ISR_TXRDY	0x02	/* Priority 4 - Transmit Empty */ 
#define ISR_MSRRDY	0x00	/* Priority 5 - MSR Interrupt */ 
#define ISR_RXXR	0x10	/* Priority 6 - Recieved Xon / Xoff */ 
#define ISR_RTSCTS	0x20	/* Priority 7 - RTS CTS State Change */ 
#define ISR_NONE	0x01

/* define bits in the line control register (LCR) */

#define LCR_WL5		0x00	/* wordsize */
#define LCR_WL6		0x01	/* wordsize */
#define LCR_WL7		0x02	/* wordsize */
#define LCR_WL8		0x03	/* wordsize */
#define LCR_SB1		0x00	/* stop bits */
#define LCR_SB2		0x04	/* stop bits */
#define LCR_PNONE	0x00	/* parity bits */
#define LCR_PODD	0x08	/* parity bits */
#define LCR_PEVEN	0x18	/* parity bits */

#define LCR_BRDE	0x80	/* enable baud rate divisors */

/* define bits in the modem control register (MCR) */

#define MCR_FORCEDTR	0x01	/* force state of DTR pin */
#define MCR_FORCERTS	0x02	/* force state of RTS pin */
#define MCR_DTRDSR	0x04	/* 1 = DTR/DSR for autoflow control, 0 = RTS/CTS */
#define MCR_LOOP	0x10	/* loop back */
#define MCR_CPS		0x80	/* clock prescaler select */

/* define bits in the IER register */

#define IER_RXINTEN	0x01
#define IER_TXINTEN	0x02
#define IER_LSINTEN	0x04
#define IER_MSINTEN	0x08

/* define bits in the enhanced function register (EFR) */

#define EFR_NONE	0x00	/* software flow control */
#define EFR_TX1		0x08
#define EFR_TX2		0x04
#define EFR_RX1		0x02
#define EFR_RX2		0x01
#define EFR_EFREN	0x10
#define EFR_SCDE	0x20	/* special character detect enable */
#define EFR_RTSEN	0x40
#define EFR_CTSEN	0x80

/* define bits in the feature control register (FTCR) */

#define FCTR_HY00	0x0	/* fifo hysteresis control */
#define FCTR_HY04	0x1
#define FCTR_HY06	0x2
#define FCTR_HY08	0x3
#define FCTR_HY08A	0x4	/* same effect as HY08 */
#define FCTR_HY16	0x5
#define FCTR_HY24	0x6
#define FCTR_HY32	0x7
#define FCTR_HY12	0xC
#define FCTR_HY20	0xD

#define FCTR_HY28	0xE
#define FCTR_HY36	0xF
#define FCTR_HY40	0x8
#define FCTR_HY44	0x9
#define FCTR_HY48	0xA
#define FCTR_HY52	0xB
#define FCTR_485EN	0x20
#define FCTR_HYTABA	0x00	/* select FIFO control table */
#define FCTR_HYTABB	0x40
#define FCTR_HYTABC	0x80
#define FCTR_HYTABD	0xC0

/* define bits in the FIFO control register (FCR) */

#define FCR_ENABLE	0x01
#define FCR_RXFRESET	0x02
#define FCR_TXFRESET	0x04

/* define bits in the line status register (LSR) */

#define LSR_RDRDY	0x01
#define LSR_ROVRN	0x02
#define LSR_RPARERR	0x04
#define LSR_RFRAMERR	0x08
#define LSR_RBREAK	0x10
#define LSR_THREMPTY	0x20
#define LSR_TSRIDLE	0x40
#define LSR_RFIFOERR	0x80

/* define bits in the REG_DEVCTR register */

#define CTR_WA		0x0001
#define CTR_RESET	0x0100
#define CTR_ENI		0x0200
#define CTR_INTEN	0x4000
#define CTR_IRQ		0x8000

/* define bits in the REG_TIMCTRL register */

#define TCTR_INTEN	0x0001
#define TCTR_SS		0x0002	/* SS means start/stop */
#define TCTR_ONESHOT	0x0004
#define TCTR_EXTCLCK	0x0008

/* External clock (generated by the Xilinx) 125 kHz for timer */

#define EXTCLOCK	125000

/* Internal clock */

#define INTCLOCK	14745600

/* IPAC module ID PROM 16-bit offsets - Data format II */

#define PROM_OFFS	0x40

#define PROM_VI		0x00
#define PROM_TA		0x01
#define PROM_4S		0x02
#define PROM_MANIDH	0x03
#define PROM_MANIDL	0x04
#define PROM_MODEL	0x05
#define PROM_REV	0x06
#define PROM_DIDL	0x08
#define PROM_DIDH	0x09
#define PROM_FLAGS	0x0A
#define PROM_BYTES	0x0B
#define PROM_CRC	0x0C
#define PROM_SPACE	0x0D

#define MANID		0x00800300
#define MODEL15		0x8515
#define MODEL16		0x8516
#define REV		0x2202

int ty8515DevCreate(TY_8515_DEV *pTy8515Dev, char *name, int rdBufSize, int wrtBufSize);
void DLS8515Tx(TY_8515_DEV *pTy8515Dev);
int DLS8515SetBaud(HY8515 *pcard, const int channel, const int baud);
int DLS8515SetOptions(HY8515 *pcard, const int channel, const int data, const int stop, const int parity, const int flow);

#endif
