/*

	VxWorks device driver for the Hytec 8515/8516 IP module

*/

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>

#include <vxLib.h>
#include <intLib.h>
#include <logLib.h>
#include <shellLib.h>
#include <taskLib.h>
#include <selectLib.h>

#include <epicsThread.h>
#include <epicsInterrupt.h>

#include "devDLS8515.h"
#include "drvIpac.h"

static int checkprom(regBase base);
static void RebootDLS8515(void *);
static void drvISR(HY8515 *pcard);
static void DLS8515TX(HY8515 *pcard);
static void DLS8515RX(HY8515 *pcard);

int DLS8515DevConfigure(const char *dev, const int baud, const int data, const int stop, const int parity, const int flow);
int DLS8516DevConfigure(const char *dev, const int baud, const int data, const int stop, const int parity, const int flow, const int tadelay, const int duplex);

/* linked list of card structures */

static ELLLIST list;

static struct timespec starttime;

int dls8515debug = -1;
epicsExportAddress(int, dls8515debug);

/*******************************************************************************
*
* Configure the 8515/8516 hardware and create to serial devices
*
* card		card number 		10 * slot + IP
* carrier	carrier number		value returned by ipacAddHy8002()
* vector	interrupt vector	192 - 255
* poll		polling period		1 - x, 0 for interrupts
* prefix	device name prefix
*
*******************************************************************************/

static int DLS85158516Configure(const int model, const int card, const int carrier, const int vector, const char *prefix)
{
	static int first = 1;
	int status, tmp;
	HY8515 *pcard;
	const int vmeslot = card / 10;
	const int ipslot = card - 10 * vmeslot;
	
/* check user supplied parameters */

	if ((vmeslot < 2) || (vmeslot > 21))
	{
		epicsStdoutPrintf("2 <= vmeslot <= 21\n");
		shellScriptAbort();
		return (ERROR);
	}
	
	if ((ipslot > 3) || (ipslot < 0))
	{
		epicsStdoutPrintf("0 <= ipslot <= 3\n");
		shellScriptAbort();
		return (ERROR);
	}

	if ((vector < 192) || (vector > 255))
	{
		epicsStdoutPrintf("192 <= vector <= 255\n");
		shellScriptAbort();
		return (ERROR);
	}

	if ((pcard = (HY8515 *) callocMustSucceed(1, sizeof(HY8515), "DLS8515Configure")) == NULL)
	{
		shellScriptAbort();
		return (ERROR);
	}
	
	pcard->card = card;
	pcard->vector = vector;

	if (first)
	{
		first = 0;

		ellInit(&list);					/* initialise linked list of serial IP modules */
	
		clock_gettime(CLOCK_REALTIME, &starttime);	/* timestamp for rate calculations */
	}

/* check card number */

	{
		HY8515 *pcard;
		
		for (pcard = (HY8515 *) ellFirst(&list); pcard; pcard = (HY8515 *) ellNext(&pcard->node))
		{
			if (card == pcard->card)
			{
				epicsStdoutPrintf("card %d already used\n", card);
				goto error;
			}
		}
	}
	
/* determine the card's address */

	pcard->reg = (regBase) ipmBaseAddr(carrier, ipslot, ipac_addrIO);

	if (pcard->reg == NULL)
	{
		epicsStdoutPrintf("%s: ipmBaseAddr failed for card %d.\n", DRIVER, card);
		goto error;
	}
	
/* probe the card */

	status = vxMemProbe((char *)(pcard->reg), READ, sizeof(short), (char *) &tmp);
			
	if (status != 0)
	{
		epicsStdoutPrintf("%s: Memory probe error for card %d.\n", DRIVER, card);
		goto error;
	}

/* check the PROM */

	if (checkprom(pcard->reg + PROM_OFFS))
	{
		goto error;
	}

	pcard->model = pcard->reg[PROM_OFFS + PROM_MODEL];

/* get the A32 bit register address for UART configuration registers */

	pcard->uart = (epicsUInt16 *) ipmBaseAddr(carrier, ipslot, ipac_addrMem);

	printf("  Regs: %p \n", pcard->uart);
	
/* add the card to our list */

	ellAdd(&list, &pcard->node);

/* initialise the card */

	pcard->reg[REG_DEVCTR] = CTR_RESET;
	pcard->reg[REG_RESET]  = 0xFF;
	pcard->reg[REG_INT0]   = (vector << 8);

	if (ipmIntConnect(carrier, ipslot, vector, (void (*)(int)) drvISR, (int) pcard) != 0)
	{
		epicsStdoutPrintf("%s: ipmIntConnect failed.\n", DRIVER);
		goto error;
	}
		
	if (ipmIrqCmd(carrier, ipslot, 0, ipac_irqEnable) != OK)
	{
		epicsStdoutPrintf("%s: ipmIrqCmd failed.\n", DRIVER);
		goto error;
	}

	if (epicsAtExit(RebootDLS8515, (void *) pcard) < 0)
	{
		epicsStdoutPrintf("%s: epicsAtExit() failed.\n", DRIVER);
		goto error;
	}	
		
/* write to all channels */

	pcard->reg[REG_DEVCTR] |= CTR_WA;

/* enable enhanced feature register */
		
	UART_WRITE_SYNC(pcard->uart, 0, UA_EFR, EFR_EFREN);
			
/* reset FIFOs then enable */
		
	UART_WRITE_SYNC(pcard->uart, 0, UA_FCR, FCR_RXFRESET | FCR_TXFRESET);
	UART_WRITE_SYNC(pcard->uart, 0, UA_FCR, FCR_ENABLE);
			
/* enable RX interrupts */
		
	UART_WRITE_SYNC(pcard->uart, 0, UA_IER, IER_RXINTEN);
			
/* set FIFO trigger level control */

	UART_WRITE_SYNC(pcard->uart, 0, UA_FCTR, FCTR_HYTABD);
		
/* FIFO trigger levels */
		
	for (tmp = 0; tmp < CHANNELS; tmp++)
	{
		pcard->dev[tmp].txtrg = TXFIFOTRIG;
		pcard->dev[tmp].rxtrg = RXFIFOTRIG;
	}
				
	UART_WRITE_SYNC(pcard->uart, 0, UA_TXTRG, pcard->dev[0].txtrg);
	UART_WRITE_SYNC(pcard->uart, 0, UA_RXTRG, pcard->dev[0].rxtrg);
		
/* disable write to all channels */
			
	pcard->reg[REG_DEVCTR] &= ~CTR_WA;

/* display uart revision */

	if (pcard->reg[REG_DVID] == 0x28)
	{
		epicsStdoutPrintf("  UART: XR16L788, rev %c\n", 'A' + pcard->reg[REG_DREV] - 1);
	}
	else
	{
		epicsStdoutPrintf("  UART: id 0x%x, rev %c\n", pcard->reg[REG_DVID], 'A' + pcard->reg[REG_DREV] - 1);
	}

/* generate device names and create devices */

	{
		char dname[DEVNAMEMAX];
		int chan;
		
		for (chan = 0; chan < CHANNELS; chan++)
		{
			epicsSnprintf(dname, DEVNAMEMAX, "/%s/%d/%d", prefix, card, chan);
			
			if (ty8515DevCreate(&(pcard->dev[chan]), dname, RXBUFFSIZE, TXBUFFSIZE) != OK)
			{
				epicsStdoutPrintf("ty8515DevCreate failed for %s\n", dname);
				goto error;
			}
		
		/* allow the TY_8515_DEV structure to point back to the HY8515 structure */
			
			pcard->dev[chan]._HY8515 = (void *) pcard;
			pcard->dev[chan].channel = chan;
			
		/* pre-configure the device */
		
			DLS8515DevConfigure(dname, 9600, 8, 2, 'N', 'N');
		}
	}

/* for triggering TX processing */

	pcard->work = epicsEventMustCreate(epicsEventEmpty);

/* mutex protection */

	pcard->mutex  = epicsMutexMustCreate();

/* message queue for RX process */
	
	pcard->queue = epicsMessageQueueCreate(FIFOSIZE * CHANNELS, sizeof(struct FIFO));

/* spawn tx/rx serial tasks */

	{
		char buffer[12];
		
		sprintf(buffer, "D%xTX%d", model, pcard->card);

		if (taskSpawn(buffer, TX_PRIO, 0, 2048, (FUNCPTR) DLS8515TX, (int) pcard, 0, 0, 0, 0, 0, 0, 0, 0, 0) == ERROR)
		{
			epicsStdoutPrintf("taskSpawn %s faild\n", buffer);
			goto error;
		}

		sprintf(buffer, "D%xRX%d", model, pcard->card);

		if (taskSpawn(buffer, RX_PRIO, 0, 2048, (FUNCPTR) DLS8515RX, (int) pcard, 0, 0, 0, 0, 0, 0, 0, 0, 0) == ERROR)
		{
			epicsStdoutPrintf("taskSpawn %s failed\n", buffer);
			goto error;
		}
	}

/* enable interrupts */

	pcard->reg[REG_DEVCTR] = CTR_INTEN;
	
	return (OK);
error:
	free(pcard);
	shellScriptAbort();
	
	return (ERROR);	
}

int DLS8515Configure(int card, int carrier, int vector, char *prefix)
{
	return (DLS85158516Configure(MODEL15, card, carrier, vector,  prefix));
}

int DLS8516Configure(int card, int carrier, int vector, char *prefix)
{
	return (DLS85158516Configure(MODEL16, card, carrier, vector,  prefix));
}

/*******************************************************************************
*
*	check that the correct IP module is installed
*
*******************************************************************************/
	
static int checkprom(regBase base)
{
	unsigned int manid;
	unsigned short model, rev, serial, flags;
	char *clock = NULL;
	
	epicsStdoutPrintf("  Base: %p\n", base);
		
	if ((base[PROM_VI] != 0x5649) || (base[PROM_TA] != 0x5441) || (base[PROM_4S] != 0x3420))
	{
		epicsStdoutPrintf("Bad VITA4 header\n");
		return (ERROR);
	}

	manid  = base[PROM_MANIDL] + (base[PROM_MANIDH] << 16);
	model  = base[PROM_MODEL];
	rev    = base[PROM_REV];
	flags  = base[PROM_FLAGS];
	serial = base[PROM_SPACE];

	if ((flags & 6) == 6)
	{
		clock = "8/32";
	}
	else
	if ((flags & 2) == 2)
	{
		clock = "8";
	}
	else
	if ((flags & 4) == 4)
	{
		clock = "4";
	}
	
	epicsStdoutPrintf(" Model: %04X\n Manid: %08X\n   Rev: %04X\nSerial: %d\n Flags: %s MHz\n", model, manid, rev, serial, clock);

	if (manid != MANID)
	{
		epicsStdoutPrintf("Bad ManID %x\n", manid);
	}
	
	if ((model != MODEL15) && (model != MODEL16))
	{
		epicsStdoutPrintf("Bad model %x\n", model);
		return (ERROR);
	}
	
	if (rev < REV)
	{
		epicsStdoutPrintf("Bad revision %x\n", rev);
	}
	
	return (OK);
}

/*******************************************************************************
*
* Main processing task to receive events from the ISR and process the active
* channels.
*
*******************************************************************************/

static void DLS8515RX(HY8515 *pcard)
{
	struct FIFO fifo;
	uint8_t rxc;
	
	while (1)
	{
		int mq, i;
		
		if ((mq = epicsMessageQueueReceive(pcard->queue, &fifo, sizeof(struct FIFO))) != sizeof(struct FIFO))
		{
			if (!(pcard->errors[CHANNELS] & MSG_MQRX))
			{
				logMsg("Error Card %d, epicsMessageQueueReceive = %d\n", pcard->card, mq, 0, 0, 0, 0);
				pcard->errors[CHANNELS] |= MSG_MQRX;
			}
			
			return;
		}
		
		rxc = UART_READ(pcard->uart, fifo.channel, UA_RXCNT);

		if (rxc > pcard->rxread[fifo.channel])
		{
			pcard->rxread[fifo.channel] = rxc;
		}

		if (dls8515debug > -1)
		{
			if (fifo.channel == dls8515debug)
			{
				logMsg("RX rxc %d\n", rxc, 0, 0, 0, 0, 0);
			}
		}
		
		for (i = 0; i < rxc; i++)
		{
			const uint8_t data = UART_READ(pcard->uart, fifo.channel, UA_RHR);
			
		/* push data into the device's receive buffer */
		
			if (tyIRd(&(pcard->dev[fifo.channel].tyDev), data) == OK)
			{
				
			/* number of bytes received */
	
				pcard->dev[fifo.channel].rxbytes++;
			}
			else
			
		/* there is no more space in the device's input buffer, we have lost characters */
		
			{
				if (!(pcard->errors[fifo.channel] & MSG_RXFULL))
				{
					logMsg("Error Card %d, channel %d. Ring Buffer full.\n", pcard->card, fifo.channel, 0, 0, 0, 0);
					pcard->errors[fifo.channel] |= MSG_RXFULL;
				}
				
				break;
			}
		}
		
	/* re-enable RX interrupts for this channel */
		
		{
			const int key = epicsInterruptLock();

			{
				const uint8_t ier = UART_READ(pcard->uart, fifo.channel, UA_IER);

				UART_WRITE_SYNC(pcard->uart, fifo.channel, UA_IER, ier | IER_RXINTEN);
			}

			epicsInterruptUnlock(key);
		}
		
	/* look for line errors */

		if (fifo.lsr & LSR_ROVRN)
		{
			if (!(pcard->errors[fifo.channel] & MSG_RXOVER))
			{
				logMsg("Error Card %d, channel %d. Overrun error\n", pcard->card, fifo.channel, 0, 0, 0, 0);
				pcard->errors[fifo.channel] |= MSG_RXOVER;
			}
		}
	}
}

/*******************************************************************************
*
* TX process triggered by an TX FIFO empty interrupt.
* 
* Write as many characters as possible, until FIFO full or no more characters.
* If there are no more characters then disable TX interrupts.
* Mutex controls simultaneous access to the UART
*
*******************************************************************************/

static void DLS8515TX(HY8515 *pcard)
{
	int channel, key;
	uint16_t txint;
	
	while (1)
	{
		do
		{
			epicsEventMustWait(pcard->work);
		}
		while (pcard->txint == 0);

/* signal that we are processing these UARTs but protect us from the interrupt processing */
	
		key = epicsInterruptLock();

			txint = pcard->txint;
		
			pcard->txint = 0;
				
		epicsInterruptUnlock(key);
	
	/* identify active channels */
	
		for (channel = 0; channel < CHANNELS; channel++)
		{
			uint8_t txcount;
			char c;
			int i;

			if (!(txint & (1 << channel))) continue;

		/* space left in TX FIFO */

			txcount = FIFOSIZE - UART_READ(pcard->uart, channel, UA_TXCNT);

		/* write as many characters as possible */
			
			for (i = 0; i < txcount; i++)
			{
				if (tyITx(&(pcard->dev[channel].tyDev), &c) == OK)
				{
					UART_WRITE(pcard->uart, channel, UA_THR, c);

					pcard->dev[channel].txbytes++;
				}
				else
				{

				/* no more characters */

					break;
				}
			}
		
		/* possibly more characters so enable interrupts */
	
			if (i == txcount)
			{
				const int key = epicsInterruptLock();

				{
					const uint8_t ier = UART_READ(pcard->uart, channel, UA_IER);

					if (!(ier & IER_TXINTEN))
					{
						UART_WRITE_SYNC(pcard->uart, channel, UA_IER, ier | IER_TXINTEN);
					}
				}

				epicsInterruptUnlock(key);
			}
			else
			
		/* no more characters so disable TX FIFO interrupts */
		
			{
				const int key = epicsInterruptLock();

				{
					const uint8_t ier = UART_READ(pcard->uart, channel, UA_IER);

					if (ier & IER_TXINTEN)
					{
						UART_WRITE_SYNC(pcard->uart, channel, UA_IER, ier & ~IER_TXINTEN);
					}
				}
				
				epicsInterruptUnlock(key);
			}
#if 0
			logMsg("TX chan %d\n", channel, 0, 0, 0, 0, 0);
#endif
		}
	}
}

/*******************************************************************************
*
* Transmit chars
*
* This function is called by the process performing the write. It only gets
* called once per write and signals the main TX process.
*
*******************************************************************************/

void DLS8515Tx(TY_8515_DEV *pTy8515Dev)
{
	HY8515 *pcard = (HY8515 *) pTy8515Dev->_HY8515;
	int key;
	
	key = epicsInterruptLock();

		pcard->txint |= (1 << pTy8515Dev->channel);
				
	epicsInterruptUnlock(key);

	epicsEventSignal(pcard->work);

	if (dls8515debug > -1)
	{
		if (pTy8515Dev->channel == dls8515debug)
		{
			logMsg("TX 0x%x\n", pcard->txint, 0, 0, 0, 0, 0);
		}
	}
}

/*******************************************************************************
*
* ISR 
* The highest pending interrupt is processed first ISR_RXRDY -> ISR_RXTO -> ISR_TXRDY
*
*******************************************************************************/

static __inline__ void RX(HY8515 *pcard, const int channel, const int itype)
{
	struct FIFO fifo;
	
/* disable RX interrupts */

	{
		const uint8_t ier = UART_READ(pcard->uart, channel, UA_IER);

		UART_WRITE_SYNC(pcard->uart, channel, UA_IER, ier & ~IER_RXINTEN);
	}

	fifo.channel = channel;

/* save LSR */

	fifo.lsr = UART_READ(pcard->uart, channel, UA_LSR);

/* send data */

	if (epicsMessageQueueTrySend(pcard->queue, &fifo, sizeof(struct FIFO)))
	{
		if (!(pcard->errors[channel] & MSG_MQTX))
		{
			logMsg("Error Card %d, channel %d. epicsMessageQueueTrySend\n", pcard->card, channel, 0, 0, 0, 0);
			pcard->errors[channel] |= MSG_MQTX;
		}
	}
	
	pcard->rxints++;
}

static void drvISR(HY8515 *pcard)
{
	const uint32_t int0 __attribute__ ((unused)) = pcard->reg[REG_INT0] & 0xff;
	const uint32_t int1 = pcard->reg[REG_INT1] & 0xff;
	const uint32_t int2 = pcard->reg[REG_INT2] & 0xff;
	const uint32_t int3 = pcard->reg[REG_INT3] & 0xff;
	const uint32_t gints = (int3 << 16) + (int2 << 8) + int1;
	int channel, tx;
#if 0
	logMsg("card %d - uart 0x%02x 0x%02x 0x%02x 0x%02x\n", pcard->card, int0, int3, int2, int1, 0);
#endif
	
/* RX Ready */

	for (channel = 0; channel < CHANNELS; channel++)
	{
		const uint8_t ints = (gints >> (3 * channel)) & 0x7;
	
		if (ints == INT_RXRDY)
		{
			RX(pcard, channel, INT_RXRDY);
		}
	}
	
/* RX Timeout */

	for (channel = 0; channel < CHANNELS; channel++)
	{
		const uint8_t ints = (gints >> (3 * channel)) & 0x7;
	
		if (ints == INT_RXTO)
		{
			RX(pcard, channel, INT_RXTO);
		}
	}	
	
/* TX empty */

	tx = 0;
	
	for (channel = 0; channel < CHANNELS; channel++)
	{
		const uint8_t ints = (gints >> (3 * channel)) & 0x7;
	
		if (ints == INT_TXRDY)
		{
			const uint8_t isr __attribute__ ((unused)) = UART_READ_SYNC(pcard->uart, channel, UA_ISR) & ISR_INTBITS;
				
		/* accumulate interrupt sources */

			pcard->txint |= (1 << channel);

			tx = 1;
		}
	}

/* trigger main process */

	if (tx)
	{
		pcard->txints++;

		epicsEventSignal(pcard->work);
	}
}

/*******************************************************************************
*
* Disable and clear interrupt, reset UARTs and IP module to keep it quiet.
*
*******************************************************************************/

static void RebootDLS8515(void *arg)
{
	HY8515 *pcard;
	
	for (pcard = (HY8515 *) ellFirst(&list); pcard; pcard = (HY8515 *) ellNext(&pcard->node))
	{
		int i;
		
		for (i = 0; i < CHANNELS; i++)
		{
			UART_WRITE_SYNC(pcard->uart, i, UA_IER, 0);
			
			UART_READ_SYNC(pcard->uart, i, UA_ISR);
		}
			
		VME_WRITE16_SYNC((void *) &(pcard->reg[REG_RESET]),  0xFF);	
		VME_WRITE16_SYNC((void *) &(pcard->reg[REG_DEVCTR]), CTR_RESET);
	}
}

/*******************************************************************************
*
* Configure UART parameters
*
* Device name format must be /<prefix>/<card number>/<channel number>
*
* data = 5, 6, 7, 8
* stop = 1, 2
* parity = N, E, O
* flow = N, S, H
*
*******************************************************************************/

static int DLS85158516DevConfigure(const int model, const char *dev, const int baud, const int data, const int stop, const int parity, const int flow, const int tadelay, const int duplex)
{
	int card, channel;
	HY8515 *pcard = (HY8515 *) ellFirst(&list);
	char *ptr, *save;
	const char *baddevice = "Bad device name %s (/<prefix>/<card>/<channel>)\n";
	
/* prefix */

	if ((ptr = epicsStrtok_r((char *) dev, "/", &save)) == NULL)
	{
		epicsStdoutPrintf(baddevice, dev);
		shellScriptAbort();
		return (ERROR);
	}

/* card */
	
	if (((ptr = epicsStrtok_r(NULL, "/", &save)) == NULL) || (sscanf(ptr, "%d", &card) != 1))
	{
		epicsStdoutPrintf(baddevice, dev);
		shellScriptAbort();
		return (ERROR);
	}
	
/* channel */
	
	if (((ptr = epicsStrtok_r(NULL, "/", &save)) == NULL) || (sscanf(ptr, "%d", &channel) != 1))
	{
		epicsStdoutPrintf(baddevice, dev);
		shellScriptAbort();
		return (ERROR);
	}

	while (pcard != NULL)
	{
		if (pcard->card == card)
		{
			break;
		}
		
		pcard = (HY8515 *) ellNext(&pcard->node);
		
		if (pcard == NULL)
		{
			epicsStdoutPrintf("Card %d not found\n", card);
			shellScriptAbort();
			return (ERROR);
		}
	}

/* check parameters */
	
	if ((channel < 0) || (channel >= CHANNELS))
	{
		epicsStdoutPrintf("Bad channel %d (0-7)\n", channel);
		shellScriptAbort();
		return (ERROR);
	}
	
	if ((data < 5) || (data > 8))
	{
		epicsStdoutPrintf("Bad data bits %d (5|6|7|8)\n", data);
		shellScriptAbort();
		return (ERROR);
	}
	
	if ((stop != 1) && (stop != 2))
	{
		epicsStdoutPrintf("Bad stop bits %d (1|2)\n", stop);
		shellScriptAbort();
		return (ERROR);
	}
	
	if ((parity != 'N') && (parity != 'E') && (parity != 'O'))
	{
		epicsStdoutPrintf("Bad parity '%c' (N|E|O)\n", parity);
		shellScriptAbort();
		return (ERROR);
	}

	if ((flow != 'N') && (flow != 'S') && (flow != 'H'))
	{
		epicsStdoutPrintf("Bad flow control '%c' (N|S|H)\n", parity);
		shellScriptAbort();
		return (ERROR);
	}

	if ((duplex != 0) && (duplex != 1))
	{
		epicsStdoutPrintf("Bad duplex %d (0|1)\n", duplex);
		shellScriptAbort();
		return (ERROR);
	}

	if ((tadelay < 0) || (tadelay > 15))
	{
		epicsStdoutPrintf("Bad rs485 delay %d (0-15)\n", tadelay);
		shellScriptAbort();
		return (ERROR);
	}

/* lookup baud rate value */

	if (DLS8515SetBaud(pcard, channel, baud) == ERROR)
	{
		epicsStdoutPrintf("Illegal baud rate %d\n", baud);
		shellScriptAbort();
		return (ERROR);
	}

	DLS8515SetOptions(pcard, channel, data, stop, parity, flow);

	if (model == MODEL16)
	{
	
	/* turn around delay */

		UART_WRITE_SYNC(pcard->uart, channel, UA_RS485, tadelay << 4);

	/* duplex = 1 for full duplex */

		pcard->reg[REG_DUPLEX] &= ~(1 << channel);
		pcard->reg[REG_DUPLEX] |= (duplex << channel);

	/* enable Auto RS485 half duplex direction control */

		if (!duplex)
		{
			UART_WRITE_SYNC(pcard->uart, channel, UA_FCTR, FCTR_HYTABD | FCTR_485EN);
		}
	}
	
	return (OK);
}

int DLS8515DevConfigure(const char *dev, const int baud, const int data, const int stop, const int parity, const int flow)
{
	return (DLS85158516DevConfigure(MODEL15, dev, baud, data, stop, parity, flow, 0, 0));
}

int DLS8516DevConfigure(const char *dev, const int baud, const int data, const int stop, const int parity, const int flow, const int tadelay, const int duplex)
{
	return (DLS85158516DevConfigure(MODEL16, dev, baud, data, stop, parity, flow, tadelay, duplex));
}

/*******************************************************************************
*
* Set baud rate for card/channel
*
* See header file for the list of currently supported baud rates
*
*******************************************************************************/

int DLS8515SetBaud(HY8515 *pcard, const int channel, const int baud)
{
	int ix;

	for (ix = 0; ix < NELEMENTS(baudTable); ix++)
	{
		if (baudTable[ix].rate == baud)
		{
			uint8_t lcr;
			
			epicsMutexLock(pcard->mutex);

				lcr = UART_READ(pcard->uart, channel, UA_LCR);

				UART_WRITE_SYNC(pcard->uart, channel, UA_LCR, lcr | LCR_BRDE);
			
				UART_WRITE_SYNC(pcard->uart, channel, UA_DLM, baudTable[ix].divisor >> 8);
				UART_WRITE_SYNC(pcard->uart, channel, UA_DLL, baudTable[ix].divisor & 0xff);

				UART_WRITE_SYNC(pcard->uart, channel, UA_LCR, lcr);

				pcard->dev[channel].baud = baud;

			epicsMutexUnlock(pcard->mutex);
				
			return (OK);
		}
	}

	return (ERROR);	
}

/*******************************************************************************
*
* Set options for card/channel
*
*******************************************************************************/

int DLS8515SetOptions(HY8515 *pcard, const int channel, const int data, const int stop, const int parity, const int flow)
{
	uint8_t lcr, efr, mcr;
	
	epicsMutexLock(pcard->mutex);

/* /4 prescaler for the input clock, RTS/CTS not DTR/DSR */

	mcr = MCR_CPS;
	
	lcr = 0x00;
	efr = EFR_NONE;
		
	pcard->dev[channel].data = data;
	pcard->dev[channel].stop = stop;
	pcard->dev[channel].flow = flow;
	pcard->dev[channel].parity = parity;
	
	switch (data)
	{
		case 5: lcr |= LCR_WL5; break;	
		case 6: lcr |= LCR_WL6; break;
		case 7: lcr |= LCR_WL7; break;		 
		case 8: lcr |= LCR_WL8; break;
		
		default: break;
	}
		
	switch (stop)
	{
		case 1: lcr |= LCR_SB1; break;
		case 2: lcr |= LCR_SB2; break;

		default: break;
	}
			
	switch (parity)
	{
		case 'N': lcr |= LCR_PNONE; break;
		case 'E': lcr |= LCR_PEVEN; break;
		case 'O': lcr |= LCR_PODD; break;

		default: break;
	}
	
	UART_WRITE_SYNC(pcard->uart, channel, UA_LCR, lcr);
#if 0

/* enable loopback for testing */
	
	mcr |= 0x10;
#endif

/* flow control */
		
	switch (flow)
	{
		case 'N':
		{
			break;
		}

	/* software flow control - Xon1 + Xoff1 */
			
		case 'S':
		{
			UART_WRITE_SYNC(pcard->uart, channel, UA_XON1,  XON);
			UART_WRITE_SYNC(pcard->uart, channel, UA_XOFF1, XOFF);

			efr =  EFR_TX1 | EFR_RX1;
						
			break;
		}

	/* full hardware flow control */
			
		case 'H':
		{
			efr = EFR_RTSEN | EFR_CTSEN;
			
		/* enable auto RTS - assert RTS to indicate that we want to send data */
		
			mcr |= MCR_FORCERTS;
			
			break;
		}
			
		default:
		{
			break;
		}
	}

/* always enable EFR extra features */

	UART_WRITE_SYNC(pcard->uart, channel, UA_EFR, efr | EFR_EFREN);
	UART_WRITE_SYNC(pcard->uart, channel, UA_MCR, mcr);
			
	epicsMutexUnlock(pcard->mutex);

	return (OK);
}

/*******************************************************************************
*
* Display UART registers
*
*******************************************************************************/

int DLS85158516Display(const int model, const int card, const int level)
{
	HY8515 *pcard;
	struct timespec endtime;
	unsigned int diff;

	for (pcard = (HY8515 *) ellFirst(&list); pcard != NULL; pcard = (HY8515 *) ellNext(&pcard->node))
	{
		int i;
		
	/* not the card we are interested in? */
	
		if ((card) && (pcard->card != card))
		{
			continue;
		}
	
	/* wrong card type */
	
		if (pcard->model != model)
		{
			continue;
		}

		epicsStdoutPrintf("\nCard: %d\n", pcard->card);

		for (i = 0; (i < CHANNELS) && (level > 0); i++)
		{
			const char *space = "   ";
			const uint8_t u_ier = UART_READ(pcard->uart, i, UA_IER);
			const uint8_t u_lcr = UART_READ(pcard->uart, i, UA_LCR);
			const uint8_t u_mcr = UART_READ(pcard->uart, i, UA_MCR);
			const uint8_t u_lsr = UART_READ(pcard->uart, i, UA_LSR);
			const uint8_t u_msr = UART_READ(pcard->uart, i, UA_MSR);
			const uint8_t u_txc = UART_READ(pcard->uart, i, UA_TXCNT);
			const uint8_t u_rxc = UART_READ(pcard->uart, i, UA_RXCNT);
			const uint8_t u_xch = UART_READ(pcard->uart, i, UA_XCHAR);
			const uint8_t u_spr = UART_READ(pcard->uart, i, UA_SPR);
			const uint8_t u_fcr = UART_READ(pcard->uart, i, UA_FCTR);
			const uint8_t u_efr = UART_READ(pcard->uart, i, UA_EFR);
			
			epicsStdoutPrintf("\n");

			epicsStdoutPrintf("IER   0x%02x%s", u_ier, space);
			epicsStdoutPrintf("LCR   0x%02X%s", u_lcr, space);
			epicsStdoutPrintf("MCR   0x%02X%s", u_mcr, space);
			epicsStdoutPrintf("LSR   0x%02X%s", u_lsr, space);
			epicsStdoutPrintf("MSR   0x%02X\n", u_msr);

			epicsStdoutPrintf("TXCNT %4d%s", u_txc, space);
			epicsStdoutPrintf("TXTRG %4d%s", pcard->dev[i].txtrg, space);
			epicsStdoutPrintf("RXCNT %4d%s", u_rxc, space);
			epicsStdoutPrintf("RXTRG %4d%s", pcard->dev[i].rxtrg, space);
			epicsStdoutPrintf("XCHAR 0x%02x\n", u_xch);

			epicsStdoutPrintf("SPR   0x%02X%s", u_spr, space);
			epicsStdoutPrintf("FCTR  0x%02X%s", u_fcr, space);
			epicsStdoutPrintf("EFR   0x%02X%s", u_efr, space);
		
		/* read baud rate - protected from simultaneous ioctl calls with semaphore */
		
			{
				uint8_t dll, dlm, lcr, efr;
				
				epicsMutexLock(pcard->mutex);
				
					lcr = UART_READ(pcard->uart, i, UA_LCR);
				
					UART_WRITE_SYNC(pcard->uart, i, UA_LCR, lcr | LCR_BRDE);
			
					dll = UART_READ(pcard->uart, i, UA_DLL);
					dlm = UART_READ(pcard->uart, i, UA_DLM);

					UART_WRITE_SYNC(pcard->uart, i, UA_LCR, lcr);
					
				epicsMutexUnlock(pcard->mutex);

					efr = UART_READ(pcard->uart, i, UA_EFR);

				{
					const uint_t  baud = (int)((INTCLOCK / 4.0) / (16.0 * dll));
					const uint8_t data = (lcr & LCR_WL8) + 5;
					const uint8_t stop = (lcr & LCR_SB2) ? 2 : 1;
					const uint8_t par  = (lcr & LCR_PEVEN) ? (((lcr & LCR_PEVEN) == LCR_PEVEN) ? 'E' : 'O') : 'N';
					const uint8_t flow = (efr & (EFR_TX1 | EFR_RX1)) ? 'S' : (efr & (EFR_RTSEN | EFR_CTSEN)) ? 'H' : 'N';
					
					epicsStdoutPrintf("DLL   0x%02X%sDLM   0x%02X  %d,%d,%d,%c,%c\n", dll, space, dlm, baud, data, stop, par, flow);
				}
			}
		}

		epicsStdoutPrintf("\n");

	/* show counts/rate */

		clock_gettime(CLOCK_REALTIME, &endtime);
	
		diff = endtime.tv_sec - starttime.tv_sec;
	
		{
			uint64_t txc[CHANNELS], rxc[CHANNELS], txt = 0, rxt = 0;
			unsigned digits = 1;
			
		/* read all channels at the same(ish) time */
		
			for (i = 0; i < CHANNELS; i++)
			{
				txt += txc[i] = pcard->dev[i].txbytes;
				rxt += rxc[i] = pcard->dev[i].rxbytes;
			}
			
		/* how many digits for width */
	
			if ((txt > 0) || (rxt > 0))
			{
				digits = (unsigned)(log10((double)((txt >= rxt) ? txt : rxt))) + 1;
			}
			
			for (i = 0; i < CHANNELS; i++)
			{
				epicsStdoutPrintf("channel %d  tx: %*lld (%5d/s)  rx: %*lld (%5d/s) FIFO = %2d\n", i, digits, txc[i], (unsigned int)(txc[i] / diff), digits, rxc[i], (unsigned int)(rxc[i] / diff), pcard->rxread[i]);
			}
			
			epicsStdoutPrintf("total      tx: %*lld (%5d/s)  rx: %*lld (%5d/s)\n", digits, txt, (unsigned int)(txt / diff), digits, rxt, (unsigned int)(rxt / diff));
		}
		
	/* interrupt count/rate */
	
		epicsStdoutPrintf("\ninterrupts: tx %lld (%d/s)  rx %lld (%d/s)\n", pcard->txints, (unsigned int)(pcard->txints / diff), pcard->rxints, (unsigned int)(pcard->rxints / diff));

	/* display error messages */
	
		for (i = 0; i < CHANNELS + 1; i++)
		{
			if (pcard->errors[i])
			{
				epicsStdoutPrintf("\nErrors:\n");
				break;
			}
		}
		
		for (i = 0; i < CHANNELS + 1; i++)
		{
			if (pcard->errors[i])
			{
				int e;
				
			/* channel 9 is unknown channel */
			
				if (i != CHANNELS)
					epicsStdoutPrintf("channel  %d: ", i);
				else
					epicsStdoutPrintf("channel -1: ");
				
				for (e = 0; e < MSG_NUM; e++)
				{
					if (pcard->errors[i] & (1 << e)) epicsStdoutPrintf("%s ", messages[e]);
				}

				epicsStdoutPrintf("\n");
			
			/* clear error messages latch so that new error messages appear */
			
				pcard->errors[i] = 0;
			}
		}
	}
	
	return (OK);
}

int DLS8515Display(const int level, const int card)
{

/* we need VX_FP_TASK for a task which calls printf */
	
	if (taskSpawn("DLS8515D", 200, VX_FP_TASK, 4096, (FUNCPTR) DLS85158516Display, MODEL15, card, level, 0, 0, 0, 0, 0, 0, 0) == ERROR)
	{
		epicsStdoutPrintf("DLS8515D: taskSpawn faild\n");
	}

	return (OK);
}

int DLS8516Display(const int level, const int card)
{

/* we need VX_FP_TASK for a task which calls printf */
	
	if (taskSpawn("DLS8516D", 200, VX_FP_TASK, 4096, (FUNCPTR) DLS85158516Display, MODEL16, card, level, 0, 0, 0, 0, 0, 0, 0) == ERROR)
	{
		epicsStdoutPrintf("DLS8515D: taskSpawn faild\n");
	}

	return (OK);
}

epicsRegisterFunction(DLS8515Configure);
epicsRegisterFunction(DLS8516Configure);
epicsRegisterFunction(DLS8515DevConfigure);
epicsRegisterFunction(DLS8516DevConfigure);
epicsRegisterFunction(DLS8515Display);
epicsRegisterFunction(DLS8516Display);
